/*    
	th_dnh_archiver - A compresser/decompresser for Danmakufu's compression format
	Copyright (C) 2010 Legend_Bird

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.x10hosting.nanohatakamachi.th_dnh.archiver;

import java.io.*;
import java.util.zip.*;

public class Decompress {
	private static byte[] MAGIC_ARCHIVE = "PACK_FILE\u0000".getBytes(); //PACK_FILE
	private static byte[] MAGIC_COMPRESS = "COMPRESS_ZIP\u0000".getBytes(); //COMPRESS_ZIP
	private static BufferedReader STDIN = new BufferedReader(new InputStreamReader(System.in));
	private Decompress(){}
	
	public static void decompress(String archive, String dest, boolean override, boolean verbose){
		try{
			File f = new File(archive);
			File d = new File(dest);
			if(!f.exists()) throw new GeneralError(f.getPath()+" dosn't exist.");
			if(!d.exists()) if(!d.mkdirs()) throw new GeneralError("Could not create "+d.getPath());
			if(!f.isFile()) throw new GeneralError(f.getPath()+" is not a file.");
			if(!d.isDirectory()) throw new GeneralError(d.getPath()+" is not a directory.");
			
			System.out.println("Decompressing "+archive+" ...");
			
			FileInputStream file_in = new FileInputStream(f);
			NDataInputStream in = new NDataInputStream(file_in);

			byte[] magic = in.readBytes(MAGIC_ARCHIVE.length);
			if(!equal(magic,MAGIC_ARCHIVE)) throw new BadArchiveException("Bad archive Magic Number ("+new String(magic)+")");
			int nofiles = in.readInt(); //If number of files > Integer.MAX_VALUE, this won't work.
									    //unlike with the file sizes however, that is unlikely to
									    //happen.
			if (verbose) System.out.println("Files: "+nofiles);
			CFile[] files = new CFile[nofiles];
			for(int i=0; i<nofiles;i++){
				files[i]=new CFile(d.getPath(),readString(in),in.readInt(),in.readInt());
				if(verbose) System.out.println(files[i]);
			}
			for(int i=0;i<files.length;i++){
				if(override&&files[i].file.exists()){
					System.out.print("Override "+files[i].file.getPath()+"? [y/n] (n) ");
					if(!STDIN.readLine().equals("y")) continue;				
				}
				
				file_in = new FileInputStream(f);file_in.skip(files[i].offset); //Do not assume continuous arrangment of file data.
				
				byte[] f_t = new byte[files[i].size];
				file_in.read(f_t); //Hopefully more efficent.
				
				FileOutputStream file_out = new FileOutputStream(files[i].file);
				
				ByteArrayInputStream f_s = new ByteArrayInputStream(f_t);
				NDataInputStream f_d = new NDataInputStream(f_s);
				
				if(equal(f_d.readBytes(MAGIC_COMPRESS.length),MAGIC_COMPRESS)) {
					byte[] decompressed = new byte[f_d.readInt()];
					
					InflaterInputStream in2 = new InflaterInputStream(f_s);
					DataInputStream in3 = new DataInputStream(in2);
					if(verbose) System.out.println("Decompressing to "+files[i].file.getPath());
					in3.readFully(decompressed, 0, decompressed.length);
					file_out.write(decompressed);
					file_out.flush();
				}
				else {
					if(verbose) System.out.println("Unpacking to "+files[i].file.getPath());
					file_out.write(f_t);
				}
			}
			System.out.println("Decompression complete.");
		} catch (GeneralError e) {
			if(e.getMessage()!=null)System.err.println(e.getMessage());
		} catch (BadArchiveException e) {
			if(e.getMessage()!=null)System.err.println("File is not a th_dnh archive: "+e.getMessage());
			else System.err.println("File is not a th_dnh archive: Error unspecificed");
		} catch (ZipException e) {
			System.err.println("File is not a th_dnh archive: Bad compression.");
			System.err.println("( Exception message: "+e.getMessage()+" )");
		} catch (FileNotFoundException e) {
			System.err.println("Unexpected Error: File not found.");
		} catch (IOException e) {
			System.err.println("Unexpected Error: Low level IO error.");
			e.printStackTrace();
		}
	}
	
	private static String readString(NDataInputStream in) throws BadArchiveException, IOException{
		int l = in.readInt();
		byte[] b = new byte[l-1];
		for(int i=0;i<l-1;i++){
			b[i]=in.readByte();
		}
		if(in.readByte()!=0) throw new BadArchiveException("String not null terminated.");
		return new String(b);
	}
	private static boolean equal(byte[] a, byte[] b){
		if(a.length!=b.length) return false;
		for(int i=0;i<a.length;i++) if(a[i]!=b[i]) return false;
		return true;
	}
	
	private static class CFile{
		private final File file;
		private final int offset;
		private final int size;
		private CFile(String path, String file, int offset, int size){
			this.file=new File(path+"/"+file);this.offset=offset;this.size=size;
		}
		public String toString(){
			return "File: "+file.getPath()+"; Offset: "+offset+"; Size: "+size;
		}
	}
	@SuppressWarnings("serial")private static class GeneralError extends Exception{public GeneralError(String string) {super(string);}}
	@SuppressWarnings("serial")private static class BadArchiveException extends Exception{public BadArchiveException(String string) {super(string);}}
}
